/*
 * Here's a sample kernel module showing the use of jprobes to dump
 * the arguments of ext4_fallocate().
 *
 * For more information on theory of operation of jprobes, see
 * Documentation/kprobes.txt
 *
 * Build and insert the kernel module as done in the kprobe example.
 * You will see the trace data in /var/log/messages and on the
 * console whenever ext4_fallocate() is invoked to create a new process.
 * (Some messages may be suppressed if syslogd is configured to
 * eliminate duplicate messages.)
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/kprobes.h>

/*
 * Jumper probe for ext4_fallocate.
 * Mirror principle enables access to arguments of the probed routine
 * from the probe handler.
 */

/* Proxy routine having the same arguments as actual ext4_fallocate() routine */
static int jext4_fallocate(struct inode *inode, int mode, loff_t offset, loff_t len)
{
	struct dentry *dentry;

	dentry = list_entry(inode->i_dentry.next, struct dentry, d_alias);
	printk(KERN_INFO
		"jprobe: ext4_fallocate by %s for inode %ld (%s), off %lld, len %lld\n",
		current->comm, inode->i_ino,
		dentry ? dentry->d_name.name : "Unknown", offset, len);
	//dump_stack();

	/* Always end with a call to jprobe_return(). */
	jprobe_return();
	return 0;
}

static struct jprobe my_jprobe = {
	.entry			= (kprobe_opcode_t *) jext4_fallocate,
	.kp = {
		.symbol_name	= "ext4_fallocate",
	},
};

static int __init jprobe_init(void)
{
	int ret;

	ret = register_jprobe(&my_jprobe);
	if (ret < 0) {
		printk(KERN_INFO "register_jprobe failed, returned %d\n", ret);
		return -1;
	}
	printk(KERN_INFO "Planted jprobe at %p, handler addr %p\n",
	       my_jprobe.kp.addr, my_jprobe.entry);
	return 0;
}

static void __exit jprobe_exit(void)
{
	unregister_jprobe(&my_jprobe);
	printk(KERN_INFO "jprobe at %p unregistered\n", my_jprobe.kp.addr);
}

module_init(jprobe_init)
module_exit(jprobe_exit)
MODULE_LICENSE("GPL");
